package gov.va.med.vistalink.rpc;

import java.util.HashMap;

/**
 * Represents the collection of parameters associated with an RPC.
 * <p>
 * J2SE Example:
 * <p>
 * <code>
 *     // request  and response objects <br>
 *     RpcRequest vReq = null; <br>
 *     RpcResponse vResp = null;
 * <p>    
 *     //The Rpc Context<br>
 *     String rpcContext = &quot;XOBV VISTALINK TESTER&quot;;
 * <p>    
 *     //The Rpc to call<br>
 *     String rpcName = &quot;XOBV TEST STRING&quot;;
 * <p>    
 *      //Construct the request object<br>
 *     vReq = RpcRequestFactory.getRpcRequest(rpcContext, rpcName);
 * <p>    
 *     //clear the params 	<br>
 *     vReq.clearParams();
 * <p>    
 *     //Set the params<br>
 *     vReq.getParams(). setParam(1, &quot;string&quot;, &quot;This is a test string!&quot;);
 * <p>    
 *     //Execute the Rpc and get the response<br>
 *     vResp = myConnection.executeRPC(vReq);
 * <p>    
 *     //Work with the response ...
 * </code>
 * 
 */
public class RpcRequestParams {

    /**
     * The node representing the params for the RpcRequest
     */
    private HashMap params = null;

    /**
     * Constructs the Rpc Params.
     */
    protected RpcRequestParams() {
        this.params = new HashMap();
    }

    /**
     * Returns the params node.
     * 
     * @return Node
     */
    protected HashMap getParams() {
        return params;
    }

    /**
     * Clears the parameters.
     *
     */
    public void clear() {
        params.clear();
    }

    /**
     * Gets the value for a parameter associated with a specified position in the parameters collection. Normally, this
     * method is not used by the client.
     * 
     * @param position
     *            parameter position the M RPC expects this parameter
     * @return Object String or Map. A return value of <b>null </b> indicates that there is no parameter for the
     *         position.
     */
    public Object getParam(int position) {
        String posStr = String.valueOf(position);
        if (params.containsKey(posStr)) {
            RpcRequestParam param = (RpcRequestParam) params.get(posStr);
            return param.getValue();
        } else {
            return null;
        }
    }

    /**
     * Sets a parameter needed by for a M RPC call.
     * <p>
     * The position argument is the parameter list position where the RPC expects to see this argument.
     * <p>
     * The type argument indicates to VistALink how the argument should be processed on the M VistA server.
     * <p>
     * Possible values are the following:
     * <ul>
     * <li>string (corresponds to 'Literal' in VA RPC Broker)
     * <li>array (corresponds to 'List' in VA RPC Broker)
     * <li>ref (corresponds to 'Reference' in VA RPC Broker)
     * </ul>
     * 
     * @param position
     *            parameter position the M RPC expects this parameter
     * @param type
     *            type of parameter corresponding to valid M RPC types
     * @param value
     *            value of parameter
     */
    public void setParam(int position, String type, Object value) {
        params.put(String.valueOf(position), new RpcRequestParam(type, value));
    }
}